(** mathObjects.BigInt.inc

   This file contains methods, used for the manipulating of PTermInteger-s,
   the big int type of mathSys.
   They work direct on the underlaying object representation.
*)

{$ifndef MATHOBJECT_INCFILE_INCLUDED}
uses mathObjects;
{$endif}

const
  DIGIT_BASE_LOG2=16;
  DIGIT_BASE=1 shl DIGIT_BASE_LOG2;
  DIGIT_MAX=TIntegerDigit(-1);
type
  TSignedIntegerDigit=longint;
{procedure TMathSystem.freeDigit(digit: PIntegerDigit);
begin
  if digit^.next <> nil then freeDigit(digit^.next);
  FreeMem(digit);
end; }

function TMathSystem.createInteger(size:longint=0;len:longint=0):PTermInteger;
begin
  getMem(result,sizeof(result^));
  FillChar(result^,sizeof(result^),0);
  result^.typ:=ttInteger;
  result^.len:=len;
  if size<=len then size:=len;
  if size<=0 then size:=1;
  SetLength(result^.digits,size)
end;

(**
  TMathSystem.digitsToStr wandelt einen BigInt einer gegebenen Basis in einen String um

  Falls die Basis <= 36 ist, wird jede Stelle durch ein Zeichen 0..9A..Z dargestellt,
  ansonsten werden die Stellen jeweils in Dezimalsystem ausgegeben und durch # getrennt

*)
function TMathSystem.digitsToStr(const digit: TTermInteger; base: TIntegerDigit): string;
var i:integer;
begin
  with digit do begin
    if base<=36 then begin
      SetLength(result,len);
      for i:=0 to len-1 do begin
        if digits[i]>=10 then result[len-i]:=chr(digits[i]-10+ord('A'))
        else result[len-i]:=chr(digits[i]+ord('0'));
      end;
    end else begin
      result:=IntToStr(digits[0]);
      for i:=1 to len-1 do
        result:=result+'#'+IntToStr(digits[i]);
    end;
    if sign then result:='-'+result;
  end;
end;

(**
  TMathSystem.digitsToStr wandelt einen String einer gegebenen Basis in einen BigInt um

  Alle Nicht alpha numerischen Zahlen werden dabei ignoriert.
  - invertiert das Vorzeichen
//  Falls die Basis <= 36 ist, wird jede Stelle durch ein Zeichen 0..9A..Z dargestellt,
  //ansonsten werden die Stellen jeweils in Dezimalsystem ausgegeben und durch # getrennt

*)
function TMathSystem.StrToDigits(const str: string; base: TIntegerDigit): PTermInteger;
var i,j:integer;
begin
  result:=createInteger(length(str),length(str));
  j:=0;
  for i:=length(str) downto 1 do begin
    case str[i] of
      '0'..'9': result.digits[j]:=ord(str[i])-ord('0');
      'A'..'Z': result.digits[j]:=ord(str[i])-ord('A')+10;
      'a'..'z': result.digits[j]:=ord(str[i])-ord('a')+10;
      '-': if j=0 then result^.sign:=not result^.sign
           else ErrorNotFound('Number');
      else dec(j);
    end;
    inc(j);
  end;
end;

function TMathSystem.convert(number:PTermInteger;baseOld,baseNew:TIntegerDigit): PTermInteger;

var oldValue: TIntegerDigit;
    pos,pos2: longint;
    firstDigit:boolean;
begin
  getMem(result,sizeof(result^));
  FillChar(result^,sizeof(result^),0);
  Result.typ:=ttInteger;
  result.len:=trunc(number.len*system.LN(baseOld)/system.LN(baseNew))+2;
  Result^.sign:=number^.sign;
  setlength(result.digits,result.len);
  fillchar(result.digits[0],length(result.digits)*sizeof(result.digits[0]),0);
  result.len:=0;
  while (number.len > 0) do begin
    //number := number div baseNew
    oldValue:=0;
    pos:=number^.len-1;
    firstDigit:=true;
    while pos>=0 do begin
      oldValue := oldValue + number^.digits[pos];
      if oldValue >= baseNew then begin
        number^.digits[pos]:=oldValue div baseNew;
        pos2:=pos;
        while (number^.digits[pos2]>=baseOld) do begin
          number^.digits[pos2]:=number^.digits[pos2]-baseOld;
          inc(pos2);
          inc(number^.digits[pos2]);
          if pos2>=number^.len then number^.len:=pos2;
        end;
        oldValue:=oldValue mod baseNew;
        firstDigit:=false;
      end else begin
        if firstDigit then dec(number^.len);
        number^.digits[pos]:=0;
      end;
      oldValue:=oldValue*baseOld;
      dec(pos);
    end;

    //digits[last] := number mod baseNew
    if result^.len>=length(result^.digits) then
      setlength(result^.digits,length(result^.digits)+50);
    result.digits[result.len]:=oldValue div baseold;
    inc(result.len);
  end;
end;

procedure TMathSystem.add(var a:PTermInteger; const b:PTermInteger);
var overflow: boolean;
    len,i:longint;

begin
  if a^.sign<>b^.sign then begin
    a^.sign:=b^.sign;
    sub(a,b);
    a^.sign:=not a^.sign;
    exit;
  end;
  len:=a^.len;
  if b^.len>len then len:=b^.len;
  if len>length(a^.digits) then
    setlength(a^.digits,len+2);
  a^.len:=len;

  overflow:=false;
  i:=0;
  while(i<min(a^.len,b^.len)) do begin
    a^.digits[i]:=a^.digits[i]+b^.digits[i];
    if overflow then begin
      inc(a^.digits[i]);
      overflow:=false;
    end;
    if a^.digits[i]>=DIGIT_BASE then begin
      a^.digits[i]:=a^.digits[i]-DIGIT_BASE;
      overflow:=true;
    end;
    inc(i);
  end;

  if overflow then begin
    dec(i);
    a^.digits[i]:=a^.digits[i]+DIGIT_BASE;
    while a^.digits[i]>DIGIT_BASE do begin
      a^.digits[i]:=a^.digits[i]-DIGIT_BASE;
      inc(i);
      if i>=len then begin
        setlength(a^.digits,len+2); //overflow possible
        a^.len:=len+1;
      end;
      inc(a^.digits[i]);
    end;
  end;
end;

procedure TMathSystem.sub(var a:PTermInteger; const b:PTermInteger);
var underflow: boolean;
    i:longint;
begin
  if a^.sign<>b^.sign then begin //In Wirklichkeit Addition?
    a^.sign:=b^.sign;
    add(a,b);
    a^.sign:=not a^.sign;
    exit;
  end;

  if (length(a.digits)<b.len) then //Speicher im Ergebnis reservieren
    setlength(a.digits,b.len);

  //Einfache Subtraktion
  underflow:=false;
  for i:=0 to min(a^.len,b^.len)-1 do begin
    TSignedIntegerDigit(a^.digits[i]):=TSignedIntegerDigit(a^.digits[i])-TSignedIntegerDigit(b^.digits[i]);
    if underflow then begin
      dec(TSignedIntegerDigit(a^.digits[i]));
      underflow:=false;
    end;
    if TSignedIntegerDigit(a^.digits[i])<0 then begin
      TSignedIntegerDigit(a^.digits[i]):=TSignedIntegerDigit(a^.digits[i])+DIGIT_BASE;
      underflow:=true;
    end;
  end;

  //berlauf korrigieren
  if (a^.len>b^.len)and underflow then begin
    //a > b, also kann er durch sptere Ziffern korrigiert werden
    i:=b^.len;
    dec(a^.digits[i]);
    while  TSignedIntegerDigit(a^.digits[i])<0 do begin
      TSignedIntegerDigit(a^.digits[i]):=TSignedIntegerDigit(a^.digits[i])+DIGIT_BASE;
      inc(i);
      dec(TSignedIntegerDigit(a^.digits[i]));
    end;
  end else if (a^.len<b^.len)or((a^.len=b^.len)and underflow) then begin
    //a < b, also liegt die Zahl nun in Komplementdarstellung vor
    //Komplementdarsetllung -> Standard
    a^.sign:=not a^.sign;
    for i:=0 to min(a^.len,b^.len) -1 do begin
      if a^.digits[i]<>0 then begin
        a^.digits[i]:=DIGIT_BASE-a^.digits[i];
        inc(a^.digits[i+1]);
      end;
    end;
    underflow:=a^.digits[min(a^.len,b^.len)-1]>b^.digits[min(a^.len,b^.len)-1];
    for i:=min(a^.len,b^.len) to b^.len-1 do begin
      if b^.digits[i]<>0 then begin
        a^.digits[i]:=b^.digits[i];
        if underflow then
          if TSignedIntegerDigit(a^.digits[i])>0 then begin
            dec(a^.digits[i]);
            underflow:=false;
          end else begin
            a^.digits[i]:=0;
            underflow:=true;
          end;
      end else if underflow then a^.digits[i]:=DIGIT_BASE-1
      else a^.digits[i]:=0;
    end;
    a^.len:=b^.len;
  end;
end;

(** Simple Multiplication
  TODO:
    - change to Karabutsa after x2 digits
    - change to FFT after x3 digits
*)
function TMathSystem.mul(const a,b:PTermInteger):PTermInteger;
var i,j:longint;
begin
  result:=createInteger(a^.len+b^.len);
  result^.sign:=a^.sign xor b^.sign;
  FillChar(result^.digits[0],length(result^.digits)*sizeof(result^.digits[0]),0);
  for i:=0 to a^.len-1 do begin
    for j:=0 to b^.len-1 do begin
      result^.Digits[i+j]:=result^.Digits[i+j]+a^.digits[ i]*b^.digits[j];
{      asm
        mov eax
      end;     }
      if result^.Digits[i+j]>=DIGIT_BASE then begin
        inc(result^.digits[i+j+1],result^.digits[i+j] div DIGIT_BASE);
        result^.digits[i+j]:=result^.digits[i+j] mod DIGIT_BASE;
        if i+j+1>=result^.len then result^.len:=i+j+2;
      end else if i+j>=result^.len then result^.len:=i+j+1;
    end;
  end;
end;

procedure TMathSystem.divMod(const a,b:PTermInteger; var result, remainder: PTermInteger);
var i,j:longint;
    underflow,overflow: TSignedIntegerDigit;
    x: TIntegerDigit;

begin
  if a<>remainder then remainder:=pterminteger(copy(pterm(a)))
  else remainder:=a;   //remainder:=pterminteger(copy(pterm(a)));
  while(b^.len>=1)and(b^.digits[b^.len-1]=0) do dec(b^.len);
  result:=createInteger(remainder^.len-b^.len+1);
  result^.sign:=remainder^.sign xor b^.sign;
  x:=0;
  for i:=remainder^.len-b^.len downto 0 do begin
    while(remainder^.len>=1)and(remainder^.digits[remainder^.len-1]=0) do dec(remainder^.len);
    //find max x with b*x*base^i <= remainder;
    //this doesn't work correctly, sometimes is x to big.
    if (remainder^.len>=2)and(remainder^.len>b^.len+i) then
      x:=((remainder^.digits[remainder^.len-1] shl DIGIT_BASE_LOG2) + remainder^.digits[remainder^.len-2]) div b^.digits[b^.len-1]
     else if remainder^.len>=1 then
      x:=remainder^.digits[remainder^.len-1] div b^.digits[b^.len-1]
     else break;
    if x>=DIGIT_BASE then x:=DIGIT_BASE-1; //necessary, since x can be tooig

    //remainder:=remainder - b*x*base^i;
    underflow:=0;
    for j:=0 to b^.len-1 do begin
      //Multiplicate b^.digits[j] with x
      //Due to an error in the Delphi unsigned multiplication, the result of two
      //multiplication with x/2 will be added
      TSignedIntegerDigit(remainder^.digits[i+j]):=TSignedIntegerDigit(remainder^.digits[i+j])-TSignedIntegerDigit((x shr 1)*b^.digits[j])-underflow;
      if TSignedIntegerDigit(remainder^.digits[i+j])<0 then begin
        underflow:=(-TSignedIntegerDigit(remainder^.digits[i+j])-1) div DIGIT_BASE+1;
        inc(TSignedIntegerDigit(remainder^.digits[i+j]),underflow*DIGIT_BASE);
      end else underflow:=0;

      if x and 1 = 1 then
       dec(TSignedIntegerDigit(remainder^.digits[i+j]),   TSignedIntegerDigit((x shr 1 + 1)*b^.digits[j]))
      else
       dec(TSignedIntegerDigit(remainder^.digits[i+j]),   TSignedIntegerDigit((x shr 1)*b^.digits[j]));
      if TSignedIntegerDigit(remainder^.digits[i+j])<0 then begin
        overflow:=(-TSignedIntegerDigit(remainder^.digits[i+j])-1) div DIGIT_BASE+1; //it is an underflow, (variable reuse as temporary)
        inc(underflow,overflow);
        inc(TSignedIntegerDigit(remainder^.digits[i+j]),overflow*DIGIT_BASE);
      end;
    end;

    //subtract overflows (when possible and necessary)
    if underflow<>0 then
      for j:=i+b^.len to remainder^.len-1 do begin //perhaps their is only one loop, but it is safer this way
        if underflow>=TSignedIntegerDigit(remainder^.digits[j]) then begin
          dec(underflow,TSignedIntegerDigit(remainder^.digits[j]));
          remainder^.digits[j]:=0;
          if underflow<=0 then break;
        end else begin
          dec(TSignedIntegerDigit(remainder^.digits[j]),underflow);
          underflow:=0;
          break;
        end;
      end;

    if underflow>0 then begin //x is too big, add b*base^i from remainder
  //    if remainder.len>=length(remainder.digits) then
//        SetLength(remainder.digits,remainder.len+1);
      while(remainder^.len>=1)and(remainder^.digits[remainder^.len-1]=0) do dec(remainder^.len);
      dec(TSignedIntegerDigit(remainder^.digits[remainder.len-1]),underflow*DIGIT_BASE);
      while (TSignedIntegerDigit(remainder^.digits[remainder.len-1])<0) do begin
        overflow:=0;
        for j:=0 to b^.len-1 do begin
          TSignedIntegerDigit(remainder^.digits[i+j]):=TSignedIntegerDigit(remainder^.digits[i+j])+TSignedIntegerDigit(b^.digits[j])+overflow;
          if TSignedIntegerDigit(remainder^.digits[i+j])>=DIGIT_BASE then begin
            overflow:=remainder^.digits[i+j] div DIGIT_BASE;
            remainder^.digits[i+j]:=remainder^.digits[i+j] mod DIGIT_BASE;
          end else overflow:=0;
        end;
        j:=b^.len+i;
        while (overflow>0)and(j<remainder^.len) do begin
          TSignedIntegerDigit(remainder^.digits[j]):=TSignedIntegerDigit(remainder^.digits[j])+overflow;
          if TSignedIntegerDigit(remainder^.digits[j])>=DIGIT_BASE then begin
            overflow:=remainder^.digits[j] div DIGIT_BASE;
            remainder^.digits[j]:=remainder^.digits[j] mod DIGIT_BASE;
          end else overflow:=0;
          inc(j);
        end;
        dec(x);
      end;
    end;

    //result:=result + x*base^i;
    result^.digits[i]:=x;
    if (x<>0) and (i>=result^.len) then result^.len:=i+1;
  end;
  if result^.len=0 then begin
    result^.len:=1;
    result^.digits[0]:=0;
  end;
  if (result^.len=1) and (result^.digits[0]=0) then
    result^.sign:=false;
end;

(**
  TMathSystem.gcd computes the greatest common divisor

  Used algorithm:
    Standard Euclidean
*)
function TMathSystem.gcd(const a,b:PTermInteger): PTermInteger;
var x,y,t,divt: PTermInteger;
begin
  x:=pterminteger(copy(pterm(a)));
  y:=pterminteger(copy(pterm(b)));
  case compare(a,b) of
    0: begin
         result:=x;
         freeterm(pterm(y));
         exit;
       end;
    1: begin
         t:=x;x:=y;y:=t;
       end;
  end;

  while (y^.len>=1) and (y^.digits[0]>0) do begin
    divMod(x,y,divt,x);
    freeterm(pterm(divt));

    t:=x;x:=y;y:=t; //exchange
  end;

  freeterm(pterm(y));
  result:=x;
end;


function TMathSystem.compare(const a,b:PTermInteger): int32s;
var i:integer;
begin
  Result:=0;
  if a^.len>b^.len then result:=1
  else if a^.len<b^.len then result:=-1
  else begin
    for i:=0 to b^.len do
      if a^.digits[i]>b^.digits[i] then result:=1
      else if a^.digits[i]<b^.digits[i] then result:=-1;
  end;
end;
{function TMathSystem.add(a,b:PIntegerDigit):PIntegerDigit;
begin
  result:=add(a,b,DIGIT_BASE);
end;
function TMathSystem.add(a,b:PIntegerDigit;base:TIntegerDigitType):PIntegerDigit;
begin

end;
function TMathSystem.sub(a,b:PIntegerDigit):PIntegerDigit;
begin
  result:=sub(a,b,DIGIT_BASE);
end;
function TMathSystem.sub(a,b:PIntegerDigit;base:TIntegerDigitType):PIntegerDigit;
begin
end;
function TMathSystem.mul(a,b:PIntegerDigit):PIntegerDigit;
begin
  result:=mul(a,b,DIGIT_BASE);
end;
function TMathSystem.divide(a,b:PIntegerDigit):PIntegerDigit;
begin
  result:=divide(a,b,DIGIT_BASE);
end;
function TMathSystem.divide(a,b:PIntegerDigit;base:TIntegerDigitType):PIntegerDigit;
begin
end;                                                                                }

