//{$define debugtree}
{$ifopt Q+}{$define checkoverflow}{$endif}

//Check:
unit mathObjects;
{
References:
  The Art of computer programmming, vol 2 seminumerical algorithms, Donald E. Knuth

  Modern Computer Algebra, Joachim von zur Gathen and Jrgen Gerhard

  Computer Algebra, 

  Algorithmen in C, R.Segdewick}

interface
uses dialogs,sysutils,math,classes,typinfo;
//  type TNumber=double;
  //     PNumber=^TNumber;
{$Z4}
const inf=1.0/0.0;
      nan=0.0/0.0;

type
  int32s=longint;
  int32u=longword;
  int64s=int64;

  TInt32UArray=array of int32u;

type
  TMathSystem=class;
  TParseOption=(poLongVariables);
  TParseOptions=set of TParseOption;

  TSingleOperator=char;
  TFunctionOperator=({foNoneReserved,}foSin,foCos,foTan,foExp,foLn,foSqrt,foAbs,foSign,foInt,foAsin,foAcos,foAtan);

  TTermType=(ttInteger,ttNumber64F,ttVariable,ttBinOp,ttPolynom,ttFuncOp,
             tt2Function); //no regular Terms
  PTerm=^TTerm;
  PTermInteger=^TTermInteger;
  PTermNumber64F=^TTermNumber64F;
  PTermVariable=^TTermVariable;
  PTermBinOp=^TTermBinOp;
  PTermPolynom=^TTermPolynom;
  PTermFuncOp=^TTermFuncOp;
  PVariable=^TVariable;
  TTerm=record
    typ:TTermType;
 //   ref: integer;
  end;

  TIntegerDigit=longword;
  TTermInteger=record
    typ:TTermType; //must be ttInteger
    len:longint;
    sign:boolean;
    digits: array of TIntegerDigit;
  end;

  TTermNumber64F=record
    typ:TTermType; //must be ttNumber64F
    number: double;
  end;

  TTermVariable=record
    typ:TTermType; //must be ttVariable
    variable: PVariable;
  end;


  TTermFuncOp=record
    typ:TTermType; //must be ttBinOp
    func: TFunctionOperator;
    param: PTerm;
  end;

  TTermBinOp=record
    typ:TTermType; //must be ttBinOp
    op: TSingleOperator; //+ or - or * or /
    t1,t2: PTerm;
  end;

  TPolynomPart=record
    factor,pot: PTerm;
  end;
  TTermPolynom=record
    typ:TTermType; //must be ttPolynom
//    ref: integer;
    variable: string;
    termCount: longint;
    parts: array[0..200000] of TPolynomPart;
  end;


  tvariableRenderInfo=record
    {constant: boolean;
    minValue64f,maxValue64f,step64f: double;}
    values: array of double;
    minColor,maxColor: longint;
    more: pointer;
  end;
  tvariable=record
    name:string;
    ref: integer;
    substitutionValue,calculationValue:PTerm;
    prev,next: PVariable;

    //for rendering
    renderInfo: tvariableRenderInfo;

    //Sonstiges
    user: pointer;
  end;


  TVariableArray=array of PVariable;
  PFunction=^TFunction;
  TFunction=record
    typ:TTermType; //must be tt2Function
//    ref: integer;
    paramVar: PVariable;
    term: PTerm;
    variables: TVariableArray;
    derivative: PFunction;
  end;

  TEnumVariables=procedure (variable: PVariable) of object;
  TVariableCreatedEvent=procedure (system: tmathSystem; variable: PVariable) of object;
  TTermOutputOptions=record
    multiplicationImplicit:boolean;
  end;
  TTermSort=(cpDifferentAndLess,cpDifferentAndGreater,cpDifferentPowers, cpEqual,cpUnknown);
  TMathSystem=class
  public
//  protected
    variables:PVariable;

    //Methods for Fourier Transformation
    procedure FFT_in_Fp(var inputFactors, outputFactors: TInt32UArray;
                        const w,p: int32u);

    //Methods applied to BigInts (PTermInteger)
    function createInteger(size:longint=0; len:longint=0):PTermInteger;
{    procedure freeDigit(digit: PIntegerDigit);}
    function digitsToStr(const digit: TTermInteger; base: TIntegerDigit): string;
    function StrToDigits(const str: string; base: TIntegerDigit): PTermInteger;
    function convert(number:PTermInteger;baseOld,baseNew:TIntegerDigit): PTermInteger;
    procedure add(var a:PTermInteger; const b:PTermInteger);  overload;
    procedure sub(var a:PTermInteger; const b:PTermInteger);  overload;
    function mul(const a,b:PTermInteger):PTermInteger;  overload;
    procedure divMod(const a,b:PTermInteger; var result, remainder: PTermInteger);  overload;
    function gcd(const a,b:PTermInteger): PTermInteger;
    function compare(const a,b:PTermInteger): int32s;
{    function add(a,b:PIntegerDigit):PIntegerDigit; overload;
    function add(a,b:PIntegerDigit;base:TIntegerDigitType):PIntegerDigit; overload;
    function sub(a,b:PIntegerDigit;base:TIntegerDigitType):PIntegerDigit; overload;
    function mul(a,b:PIntegerDigit):PIntegerDigit; overload;
    function mul(a,b:PIntegerDigit;base:TIntegerDigitType):PIntegerDigit; overload;
    function divide(a,b:PIntegerDigit):PIntegerDigit; overload;
    function divide(a,b:PIntegerDigit;base:TIntegerDigitType):PIntegerDigit; overload;}
    function copy(const term: pterm):pterm;
   public
    termOutputOptions: TTermOutputOptions;
    onVariableCreatedEvent: TVariableCreatedEvent;

    function createFunction(const param:string; const term: PTerm):PFunction; overload;
    function createFunction(const param:PVariable; const term: PTerm):PFunction; overload;

    function createNumber(n: double): PTerm;
    function createVariable(n: string): PTerm;
    function createFunctionOperator(const op:TFunctionOperator; const param: PTerm):PTerm;


    function findVariable(name:string):PVariable;
    procedure setVariableValue(const variable:PVariable; const value: double); overload;
    procedure setVariableValue(const name:string; const value: double);        overload;
                                                                   
    function add(a,b:PTerm): PTerm;       overload;
    function sub(a,b:pterm): PTerm;       overload;
    function mul_simple(a,b:PTerm): PTerm;        //non-destructive
    function mul(a,b:PTerm): PTerm;       overload;
    function divide(a,b:PTerm): PTerm;    overload;
    function power_simple(a,b:PTerm): PTerm; //non-destructive
    function power(a,b:PTerm): PTerm;
    function exp(t: PTerm): PTerm;
    function ln(t: PTerm): PTerm;

//    function compareTermsSimple(const a,b: PTerm): TSimpleTermRelation;

    function parse(term: string; const parseOptions: TParseOptions):pterm; //non-destructive
    function printTerm(term: PTerm): string;                               //non-destructive

    function approximate(term: PTerm): double;//non-destructive

//    procedure simplify(var term: pterm);                                   //destructive
//    procedure simplify(var term: pterm);                                   //destructive
    function derivate(const func: PFunction):PFunction;overload; //non-destructive
    function derivate(const term:pterm; const variableName: string):PTerm;overload; //non-destructive
    function derivate(const term:pterm; const variable: PVariable):PTerm;overload; //non-destructive

    function searchUsedVariables(const term: pterm):TVariableArray;
    function containsVariabale(const term:pterm; const variable: PVariable):boolean;

    procedure enumVariables(const enumFunc: TEnumVariables);

    procedure FreeTerm(term: pterm);                                      //destructive

    constructor create;
    destructor destroy;override;
 end;

 function isNumber(const s:string):boolean;
 function parseNumber(const str:string): double;
implementation
  type
    TOperator=record
      op:char;
      bracketCount: longint;
    end;



  procedure error(error: string);
  begin
    raise Exception.create(error)
  end;

  procedure errorNotFound(expected: string;found: string='');
  begin
    if found<>'' then  raise Exception.create('"'+expected + '" expected, but "'+found + '" found')
    else raise Exception.create('"'+expected + '" expected, but not found')
  end;

  procedure errorEnd(expected: string);
  begin
    raise Exception.create(expected + ' expected, but nothing more found');
  end;

  {$define MATHOBJECT_INCFILE_INCLUDED}
  {$include 'mathObjects.FFT.inc'}
  {$include 'mathObjects.BigInt.inc'}
  {$include 'mathObjects.Operators.inc'}

  function parseNumber(const str:string): double;
  var signum:integer;
      start,pos,len:integer;
  begin
    pos:=1;
    len:=length(str);
    while (str[pos]=' ') do inc(pos);
    if pos>len then errorEnd('Number');

    //Vorzeichen parsen
    signum:=1;
    if str[pos]='-' then begin
      signum:=-1;
      inc(pos);
    end else if str[pos]='+' then inc(pos);

    while str[pos]=' ' do inc(pos);
    if pos>len then errorEnd('Number');

    if not (str[pos] in ['0'..'9']) then errorNotFound('Number');
    start:=pos;
    while (pos<=len) and (str[pos] in ['0'..'9']) do
      inc(pos);

    if (pos<=len) and (str[pos]='.') then begin
      inc(pos);
      while (pos<=len)and(str[pos] in ['0'..'9']) do
        inc(pos);
      Result:=StrToFloat(copy(str,start,pos-start))*signum;
    end else begin
      Result:=strtoint(copy(str,start,pos-start))*signum;
    end;
  end;

  function isNumber(const s:string):boolean;
  begin
    result:=(s<>'')and(s[1] in ['0'..'9']);
    if result then
      try
        parseNumber(s);
      except
        result:=false;
      end;
  end;



  function getPriority(const op:TSingleOperator; const bracketCount: longint=0): integer; overload;
  begin
    case op of
      '|': result:=0;
      '+','-': Result:=1;
      '*',':','/': Result:=2;
      'f': result:=3;         //Funktion
      '^': result:=4;
      '_': Result:=5;         //Vorzeichenwechsel
      else begin
             Result:=0;
             errorNotFound('Operator');
           end;
    end;
    Result:=result+10*bracketCount;
  end;


  function getPriority(const op:TOperator): integer; overload;
  begin
    Result:=getPriority(op.op,op.bracketCount)
  end;

  constructor TMathSystem.create;
  begin
    variables:=nil;
    termOutputOptions.multiplicationImplicit:=true;
  end;
  destructor TMathSystem.destroy;
  begin
    inherited;
  end;


  function TMathSystem.copy(const term: pterm):pterm;
    function copyOne(const term:pterm):pterm;
    var size: longint;
    begin
{        inc(term^.ref);
      result:=term;}
      case term^.typ of
        ttInteger: size:=sizeof(TTermInteger);
        ttNumber64F: size:=sizeof(TTermNumber64F);
        ttVariable: size:=sizeof(TTermVariable);
        ttBinOp: size:=sizeof(TTermBinOp);
        ttPolynom: size:=sizeof(TTermPolynom);
        ttFuncOp: size:=sizeof(TTermFuncOp);
        tt2Function: size:=sizeof(TFunction);
        else begin
          size:=0;
          error('unknown type (copyone)');
        end;
      end;           //size=plongint(longint(term)-4)^
      GetMem(result,size);
      move(term^,result^,size);
    end;
  begin
    Result:=copyOne(term);
    case term^.typ of
      ttVariable: inc(PTermVariable(result)^.variable^.ref);
      ttBinOp: with PTermBinOp(result)^ do begin
                 t1:=copy(t1);
                 t2:=copy(t2);
               end;
     ttFuncOp: with PTermFuncOp(result)^ do param:=copy(param);
     ttInteger: SetLength(PTermInteger(result)^.digits,length(PTermInteger(result)^.digits));
    end;
  end;

  procedure TMathSystem.FreeTerm(term: pterm);
//  var t: PIntegerDigit;
  begin
    if term=nil then raise exception.create('tried to free nil');
    case term^.typ of
      ttVariable: begin
//                    if PTermVariable(term)^.power<>nil then FreeTerm(PTermVariable(term)^.power);
                    dec(PTermVariable(term)^.variable^.ref);
                  end;
      ttInteger: begin
{                   while (PTermInteger(term)^.number.next<>nil) do begin
                     t:=PTermInteger(term)^.number.next;
                     PTermInteger(term)^.number.next:=t^.next;
                     FreeMem(t);
                   end;]   }
                   ptermInteger(term)^.digits:=nil;
                 end;
      ttFuncOp: FreeTerm(PTermFuncOp(term)^.param);
      ttBinOp:  begin
                  FreeTerm(PTermBinOp(term)^.t1);
                  if PTermBinOp(term)^.t2<>nil then
                    FreeTerm(PTermBinOp(term)^.t2);
                end;
    end;
//    dec(term^.ref);
  //  if term^.ref=0 then
      FreeMem(term);
  end;


  function TMathSystem.createFunction(const param:string; const term: PTerm):PFunction;
  begin
    result:=createFunction(findVariable(param),term);
  end;
  function TMathSystem.createFunction(const param:PVariable; const term: PTerm):PFunction;
  begin
    GetMem(result,sizeof(TFunction));
    FillChar(result^,sizeof(TFunction),0);
    Result^.typ:=tt2Function;
//    Result^.ref:=1;
    Result^.paramVar:=param;
    Result^.term:=term;
    Result^.variables:=searchUsedVariables(term);
    Result^.derivative:=nil;
  end;

  function TMathSystem.createNumber(n: double): PTerm;
  begin
    GetMem(result,sizeof(ttermNumber64F));
    Result^.typ:=ttNumber64F;
//    Result^.ref:=1;
    ptermNumber64F(result)^.number:=n;
  end;

  function TMathSystem.createVariable(n: string): PTerm;
  var i:integer;
      number: boolean;
      newVar,curVar: PVariable;
  begin
    number:=true;
    for i:=1 to length(n) do
      if not (n[i] in ['0'..'9']) then begin
        number:=false;
        break;
      end;
    if number then Result:=createNumber(StrToFloat(n))
    else begin
      GetMem(result,sizeof(ttermvariable));
      fillchar(result^,sizeof(ttermVariable),0);
      with (PTermVariable(result))^ do begin
        typ:=ttVariable;
//        ref:=1;
//        power:=createNumber(1);
      end;

      n:=LowerCase(n);
      curVar:=variables;
      while (curVar<>nil) do begin
        if curVar^.name=n then begin
          inc(curvar^.ref);
          (PTermVariable(result))^.variable:=curVar;
          exit;
        end;
        curvar:=curvar.next;
      end;

      new(newVar);
      FillChar(newVar^,sizeof(newVar^),0);
      newVar^.name:=n;
      newVar^.ref:=1;
      if n='pi' then begin
        newVar^.substitutionValue:=mul(createNumber(2),createFunctionOperator(foAcos,createNumber(0)));
        newVar^.calculationValue:=mul(createNumber(2),createFunctionOperator(foAcos,createNumber(0)));
      end else begin
        newVar^.substitutionValue:=nil;
        newVar^.calculationValue:=nil;
      end;
      newVar^.prev:=nil;
      newVar^.next:=variables;
      (PTermVariable(result))^.variable:=newVar;
      if variables<>nil then variables^.prev:=newVar;
      variables:=newVar;

      if assigned(onVariableCreatedEvent) then
        onVariableCreatedEvent(self,newVar);
    end;
  end;

  function TMathSystem.createFunctionOperator(const op:TFunctionOperator; const param: PTerm):PTerm;
  var funcOp:PTermFuncOp;
  begin
    GetMem(funcOp,sizeof(TTermFuncOp));
    funcOp^.typ:=ttFuncOp;
    funcOp^.func:=op;
    funcOp^.param:=param;
    result:=pterm(funcop);
  end;



  function TMathSystem.findVariable(name:string):PVariable;
  begin
    name:=lowercase(name);
    Result:=variables;
    while (Result<>nil) do begin
      if Result^.name=name then exit;
      result:=result.next;
    end;
  end;
  procedure TMathSystem.setVariableValue(const variable:PVariable; const value: double);
  begin
    if variable^.calculationValue = nil then variable^.calculationValue:=createNumber(value)
    else if variable^.calculationValue^.typ=ttNumber64f then
      PTermNumber64F(variable^.calculationValue)^.number:=value
    else begin
      FreeTerm(variable^.calculationValue);
      variable^.calculationValue:=createNumber(value);
    end;
  end;
  procedure TMathSystem.setVariableValue(const name:string; const value: double);
  begin
    setVariableValue(findVariable(name),value);
  end;


  function TMathSystem.parse(term: string; const parseOptions: TParseOptions):pterm;
  var pos,len:integer;

    function parseNumber(): PTerm;
    var signum:integer;
        start:integer;
    begin
      while (term[pos]=' ') do inc(pos);
      if pos>len then errorEnd('Number');

      //Vorzeichen parsen
      signum:=1;
      if term[pos]='-' then begin
        signum:=-1;
        inc(pos);
      end else if term[pos]='+' then inc(pos);

      while term[pos]=' ' do inc(pos);
      if pos>len then errorEnd('Number');

      if not (term[pos] in ['0'..'9']) then errorNotFound('Number');
      start:=pos;
      while term[pos] in ['0'..'9'] do
        inc(pos);

      if term[pos]='.' then begin
        inc(pos);
        while (term[pos] in ['0'..'9']) do
          inc(pos);
        GetMem(result,sizeof(ttermNumber64F));
        Result^.typ:=ttNumber64F;
        ptermNumber64F(result)^.number:=StrToFloat(system.copy(term,start,pos-start))*signum;
      end else begin
        GetMem(result,sizeof(ttermNumber64F));
        Result^.typ:=ttNumber64F;
        ptermNumber64F(result)^.number:=strtoint(system.copy(term,start,pos-start))*signum;
      end;
    end;



  var termstack: array of pterm;
      operatorStack:array of TOperator;
      functionStack:array of record
                      func:TFunctionOperator;
                      branckets: longint;
                    end;
    //  bracketStack: array of longint;
      nsp,osp,fsp: longint;
      lookBackPos: longint;
      lookBackStr: string;
      i:longint;
      newOp:TSingleOperator;
      newFunc:TFunctionOperator;
      bracketCount:longint;

      expectOperator: boolean;

    procedure calculateLastOperator();
    var a,b,res: PTerm;
    begin
      if (operatorStack[osp].op='f')or(operatorStack[osp].op='|') then begin //Funktion
        if nsp<0 then errorEnd('Number');
        termStack[nsp]:=createFunctionOperator(functionStack[fsp].func,termStack[nsp]);
        dec(fsp);
      end else begin
        if nsp<1 then errorEnd('Number');
        a:=termStack[nsp-1];
        b:=termStack[nsp];
        case operatorStack[osp].op of
           '+': res:=add(a,b);
           '-','_': res:=sub(a,b);
           '*': res:=mul(a,b);
           ':','/': res:=divide(a,b);
           '^': res:=power(a,b);
           else begin
                  res:=nil; //i hate this warnings!
                  errorNotFound('Operator');
                end;
        end;
        termStack[nsp-1]:=res;
        dec(nsp);
      end;
      dec(osp);
    end;

  begin
    SetLength(termStack,length(term)); nsp:=-1;
    SetLength(operatorStack,length(term)); osp:=-1;
    SetLength(functionStack,length(term)); fsp:=-1;
    len:=length(term);
    term:=term+#0;
    expectOperator:=false;

    bracketCount:=0;
    pos:=1;
    while pos<=len do begin
      if term[pos] <> ' ' then begin
        if expectOperator then begin
          case term[pos] of
            '+','-','*',':','/','^', //mgliche Operatoren
            '(',                     //ausgelassenes Mal
            'a'..'z','A'..'Z','_',   //Variable (=>mal)
            '0'..'9',                //Number (=>mal)
            '|':                     //Betrag (Ende)
                 begin
                   if (fsp>-1) and (functionStack[fsp].func=foAbs)and(term[pos]='|') then begin
                     dec(bracketCount);
                     while (osp>=0)and(operatorStack[osp].bracketCount>bracketCount) do
                       calculateLastOperator;
                   end else begin
                     newop:=term[pos];
                     if not (newop in ['+','-','*',':','/','^']) then
                       newOp:='*';
                     while (osp>=0)and(getPriority(newop,bracketCount)<=getPriority(operatorStack[osp])) do
                        calculateLastOperator;

                     inc(osp);
                     operatorStack[osp].op:=newOp;
                     operatorStack[osp].bracketCount:=bracketCount;
                     expectOperator:=false;

                     case term[pos] of
                       '(': inc(bracketCount);
                       '0'..'9','a'..'z','A'..'Z','_': dec(pos); //try again, looking for numbers
                       '|': begin
                              dec(pos);
                            end;
                     END;
                   end;
                 end;
            ')': begin //klammer schlieen und ergebnis berechnen
                   dec(bracketCount);
                   while (osp>=0)and(operatorStack[osp].bracketCount>bracketCount) do
                     calculateLastOperator;
                 end;
            else errorNotFound('Operator');
          end;
        end else begin
          case term[pos] of
            '0'..'9', '+':
                 begin
                   inc(nsp);
                   termStack[nsp]:=parseNumber;
                   dec(pos);
                   expectOperator:=true;
                 end;
            '-': begin  //read -2 as 0-2
                   inc(nsp);
                   termStack[nsp]:=createNumber(0);
                   inc(osp);
                   operatorStack[osp].op:='_';
                   operatorStack[osp].bracketCount:=bracketCount;
                 end;
            '(': inc(bracketCount);
            '|': begin
                   inc(fsp);inc(osp);
                   Inc(bracketCount);
                   functionStack[fsp].func:=foAbs;
                   functionStack[fsp].branckets:=bracketCount;
                   operatorStack[osp].op:='|';
                   operatorStack[osp].bracketCount:=bracketCount;
                 end;
            'a'..'z','A'..'Z','_': begin
                                     lookBackPos:=pos;
                                     if (poLongVariables in parseOptions ) then
                                       while (term[pos] in ['a'..'z','A'..'Z','0'..'9','_']) do inc(pos)
                                      else
                                       while (term[pos] in ['a'..'z','A'..'Z','_']) do inc(pos);
                                     inc(nsp);
                                     lookBackStr:=lowerCase(system.copy(term,lookBackPos,pos-lookBackPos));
                                     newFunc:=TfunctionOperator(GetEnumValue(typeinfo(TFunctionOperator),'fo'+lookBackStr));

                                     expectOperator:=true;
                                     if newFunc<>TfunctionOperator(-1) then begin
                                       inc(fsp);inc(osp);
                                       dec(nsp);
                                       functionStack[fsp].func:=newFunc;
                                       functionStack[fsp].branckets:=bracketCount;
                                       operatorStack[osp].op:='f';
                                       operatorStack[osp].bracketCount:=bracketCount;
                                       expectOperator:=false;
                                       continue;
                                     end else if (length(lookBackStr)>1) and not (poLongVariables in parseOptions )
                                                 and (lookBackStr<>'pi') then begin
                                       termStack[nsp]:=createVariable(lookBackStr[1]);
                                       while (osp>=0)and(getPriority('*',bracketCount)<=getPriority(operatorStack[osp])) do
                                         calculateLastOperator;
                                       for i:=2 to length(lookBackStr) do begin
                                         inc(osp);
                                         operatorStack[osp].op:='*';
                                         operatorStack[osp].bracketCount:=bracketCount;
                                         inc(nsp);
                                         termStack[nsp]:=createVariable(lookBackStr[i]);
                                       end;
                                     end else termStack[nsp]:=createVariable(lookBackStr);

                                     dec(pos);
                                   end;
            else errorNotFound('Value');
          end;

          //Klammerlose Funktionen erkennen (z.B.: tan x statt tan(x))
          while (fsp>-1) and (functionStack[fsp].branckets=bracketCount)
                         and (functionStack[fsp].func<>foAbs)
                         and (operatorStack[osp].op='f') do calculateLastOperator;
        end;
      end;
      inc(pos);
    end;

    while (osp>=0) do
      calculateLastOperator;
    Result:=termStack[0];
  end;


  function TMathSystem.printTerm(term: PTerm): string;
  var l,r: string;
  begin
    if term=nil then begin
      Result:='<error: nil>';
      exit;
    end;
    case term^.typ of
      ttNumber64F: Result:=FloatToStr(PTermNumber64F(term)^.number);
      ttVariable:  Result:=PTermVariable(term)^.variable^.name;
      ttBinOp: begin
                 l:=printTerm(PTermBinOp(term)^.t1);
                 if PTermBinOp(term)^.t1^.typ=ttBinOp then
                   if (getPriority(PTermBinOp(PTermBinOp(term)^.t1)^.op)<getPriority(PTermBinOp(term)^.op)) then
                     l:='('+l+')';

                 r:=printTerm(PTermBinOp(term)^.t2);
                 if PTermBinOp(term)^.t2^.typ=ttBinOp then
                   if ((PTermBinOp(term)^.op in ['-','/'])and(getPriority(PTermBinOp(PTermBinOp(term)^.t2)^.op)=getPriority(PTermBinOp(term)^.op))) or
                      (getPriority(PTermBinOp(PTermBinOp(term)^.t2)^.op)<getPriority(PTermBinOp(term)^.op)) then
                     r:='('+r+')';

                 if PTermBinOp(term)^.op='*' then begin
                   if ((r[1] in ['0'..'9'])and(l[length(l)] in ['0'..'9'])) or ((PTermBinOp(term)^.t1^.typ=ttBinOp)and(PTermBinOp(PTermBinOp(term)^.t1)^.op='/')) then Result:=l+'*'+r
                   else Result:=l+r;
                 end else if PTermBinOp(term)^.op='-' then begin
                   if l='0' then Result:='-'+r
                   else Result:=l+'-'+r;
                 end else Result:=l+PTermBinOp(term)^.op+r
               end;

      ttFuncOp: begin
                  Result:=GetEnumName(typeinfo(TfunctionOperator),ord(PTermFuncOp(term)^.func));
                  delete(result,1,2);
                  Result:=lowercase(Result)+'('+printTerm(PTermFuncOp(term)^.param)+')';
                end;

      tt2Function: Result:=printTerm(PFunction(term)^.term);

    end;
  end;

  function TMathSystem.approximate(term:  PTerm): double;
  begin
    Result:=0;
    case term^.typ of
       ttNumber64F: Result:=PTermNumber64F(term)^.number;
       ttVariable: begin
                     if PTermVariable(term)^.variable^.calculationValue <> nil then
                      Result:=approximate(PTermVariable(term)^.variable^.calculationValue)
                     else if PTermVariable(term)^.variable^.substitutionValue <> nil then
                      Result:=approximate(PTermVariable(term)^.variable^.substitutionValue)
                     else error('No value set for: '+PTermVariable(term)^.variable^.name);
                   end;
       ttFuncOp: case PTermFuncOp(term)^.func of
                   foSin: Result:=sin(approximate(PTermFuncOp(term)^.param));
                   foCos: Result:=cos(approximate(PTermFuncOp(term)^.param));
                   foTan: Result:=tan(approximate(PTermFuncOp(term)^.param));
                   foExp: Result:=system.exp(approximate(PTermFuncOp(term)^.param));
                   foLn: Result:=system.ln(approximate(PTermFuncOp(term)^.param));
                   foSqrt: Result:=sqrt(approximate(PTermFuncOp(term)^.param));
                   foAbs: Result:=abs(approximate(PTermFuncOp(term)^.param));
                   foSign: begin
                             Result:=approximate(PTermFuncOp(term)^.param);
                             if result<0 then Result:=-1
                             else if Result>0 then Result:=1;
                           end;
                   foInt: Result:=trunc(approximate(PTermFuncOp(term)^.param));
                   foAsin: Result:=arcsin(approximate(PTermFuncOp(term)^.param));
                   foAcos: Result:=arccos(approximate(PTermFuncOp(term)^.param));
                   foAtan: Result:=arctan(approximate(PTermFuncOp(term)^.param));
                 end ;
       ttBinOp: case PTermBinOp(term)^.op of
                  '+': result:=approximate(PTermBinOp(term)^.t1)+
                               approximate(PTermBinOp(term)^.t2);
                  '-': result:=approximate(PTermBinOp(term)^.t1)-
                               approximate(PTermBinOp(term)^.t2);
                  '*': result:=approximate(PTermBinOp(term)^.t1)*
                               approximate(PTermBinOp(term)^.t2);
               ':','/':result:=approximate(PTermBinOp(term)^.t1)/
                               approximate(PTermBinOp(term)^.t2);
                  '^': result:=math.power(approximate(PTermBinOp(term)^.t1),
                               approximate(PTermBinOp(term)^.t2));
       end;
    end;
  end;

  function TMathSystem.derivate(const term:pterm; const variableName: string):PTerm; //non-destructive
  begin
    result:=derivate(term,findVariable(variableName));
  end;

  function TMathSystem.derivate(const func: PFunction):PFunction; //non-destructive
  begin
    result:=createFunction(func^.paramVar,derivate(func^.term,func^.paramVar));
  end;

  function TMathSystem.derivate(const term:pterm; const variable: PVariable):PTerm;
  var a,b:PTerm;
  begin
    Result:=nil;
    case term^.typ of
      ttNumber64F: Result:=createNumber(0);
      ttVariable:  if PTermVariable(term)^.variable=variable then begin
                     result:=createNumber(1);
                     exit;
                   end else Result:=createNumber(0);
      ttFuncOp:    begin
                     a:=PTermFuncOp(term)^.param;
                     case PTermFuncOp(term)^.func of
                       foSin: begin
                                Result:=copy(term);
                                PTermFuncOp(result)^.func:=foCos;
                              end;
                       foCos: begin
                                Result:=copy(term);
                                PTermFuncOp(result)^.func:=foSin;
                                Result:=sub(createNumber(0),result);
                              end;
                       foTan: begin
                                result:=copy(term);
                                PTermFuncOp(result)^.func:=foCos;
                                Result:=divide(createNumber(1),mul(result,copy(term)));
                              end;
                       foExp: Result:=copy(term);
                       foLn: Result:=divide(createNumber(1),a);
                       foSqrt: Result:=divide(createNumber(0.5),copy(term));
                       foAbs: begin
                                Result:=copy(term);
                                PTermFuncOp(result)^.func:=foSign;
                              end;
                       foSign: begin
                                Result:=createNumber(0);
                                exit;
                              end;
                       foInt: begin
                                Result:=createNumber(0);
                                exit;
                              end;
                       foAsin: begin
                                 Result:=divide(createNumber(-1),
                                                createFunctionOperator(foSqrt,sub(createNumber(1),mul_simple(copy(a),copy(a)))));
                               end;
                       foAcos: begin
                                 Result:=divide(createNumber(1),
                                                createFunctionOperator(foSqrt,sub(createNumber(1),mul_simple(copy(a),copy(a)))));
                               end;
                       foAtan: begin
                                 Result:=divide(createNumber(1),
                                                add(createNumber(1),mul_simple(copy(a),copy(a))));
                               end;
                       else error('unknown function');

                     end;
                     Result:=mul(Result,derivate(a,variable));
                   end;
      ttBinOp:     begin
                     a:=PTermBinOp(term)^.t1;
                     b:=PTermBinOp(term)^.t2;
                     case PTermBinOp(term)^.op of
                       '+': Result:=add(derivate(a,variable),derivate(b,variable));
                       '-': Result:=sub(derivate(a,variable),derivate(b,variable));
                       '*': Result:=add(mul(derivate(a,variable),copy(b)),mul(copy(a),derivate(b,variable)));
                       '/': Result:=divide(sub(mul(derivate(a,variable),copy(b)),mul(copy(a),derivate(b,variable))),mul(copy(b),copy(b)));
                       '^': begin
                              if a^.typ=ttNumber64F then Result:=mul(mul(ln(copy(a)),copy(term)),derivate(b,variable))
                              else if (not containsVariabale(b,variable))and(a^.typ=ttVariable)and(PTermVariable(a)^.variable=variable) then Result:=mul(copy(b),power(copy(a),sub(copy(b),createNumber(1))))
                              else Result:=mul(exp(mul(ln(copy(a)),copy(b))),derivate(mul(ln(copy(a)),copy(b)),variable));
                            end;
                     end;
                   end;
    end;
  end;

  function TMathSystem.searchUsedVariables(const term: pterm):TVariableArray;
    procedure mergeToResult(const vars:TVariableArray);
    var i,j:integer;
        f: boolean;
    begin
      for i:=0 to high(vars) do begin
        f:=false;
        for j:=0 to high(result) do
          if result[j]=vars[i] then begin
            f:=true;
            break;
          end;
        if not f then begin
          SetLength(result,length(result)+1);
          result[high(result)]:=vars[i];
        end;
      end;
    end;
  begin
    setlength(Result,0);
    case term^.typ of
//      ttNumber64F: ;
      ttVariable: begin
                    setlength(Result,1);
                    Result[0]:=PTermVariable(term)^.variable;
                  end;
      ttBinOp:    begin
                    mergeToResult(searchUsedVariables(PTermBinOp(term)^.t1));
                    mergeToResult(searchUsedVariables(PTermBinOp(term)^.t2));
                  end;
      ttFuncOp:         Result:=searchUsedVariables(PTermFuncOp(term)^.param);
      tt2Function:      Result:=PFunction(term)^.variables;
    end;
  end;
  function TMathSystem.containsVariabale(const term:pterm; const variable: PVariable):boolean;
  begin
    Result:=false;
    case term^.typ of
      ttNumber64f:;
      ttVariable:  Result :=PTermVariable(term)^.variable=variable;
      ttFuncOp:    Result :=containsVariabale(PTermFuncOp(term)^.param,variable);
      ttBinOp:     Result :=containsVariabale(PTermBinOp(term)^.t1,variable) or
                            containsVariabale(PTermBinOp(term)^.t2,variable);
      else error('unknown type (containsVariabale)');
    end;
  end;

  procedure TMathSystem.enumVariables(const enumFunc: TEnumVariables);
  var v: PVariable;
  begin
    v:=variables;
    while (v<>nil) do begin
      enumFunc(v);
      v:=v^.next;
    end;
  end;
begin
  DecimalSeparator:='.';
end.            //      x^x = exp(ln(x)*x) = exp(x) o (ln(x) * x) = exp(ln(x) * x)




